var EventDispatcher, PathMap;
var __slice = Array.prototype.slice;
EventDispatcher = require('./EventDispatcher');
PathMap = function() {
  this.count = 0;
  this.ids = {};
  this.paths = {};
  this.arrays = {};
};
PathMap.prototype = {
  id: function(path) {
    var id;
    return this.ids[path] || (this.paths[id = ++this.count] = path, this.indexArray(path, id), this.ids[path] = id);
  },
  indexArray: function(path, id) {
    var arr, index, match, name, remainder, set;
    if (match = /^(.+)\.(\d+)(\..+|$)/.exec(path)) {
      name = match[1];
      index = +match[2];
      remainder = match[3];
      arr = this.arrays[name] || (this.arrays[name] = []);
      set = arr[index] || (arr[index] = {});
      return set[id] = remainder;
    }
  },
  init: function(count, ids) {
    var id, path, paths, _ref, _results;
    this.count = count;
    this.ids = ids;
    paths = this.paths;
    _ref = this.ids;
    _results = [];
    for (path in _ref) {
      id = _ref[path];
      paths[id] = path;
      _results.push(this.indexArray(path, id));
    }
    return _results;
  }
};
exports.init = function(model, dom, view) {
  var event, events, incrementMapItems, insert, pathMap, refIndex, remove, _fn, _i, _len, _ref;
  pathMap = model.__pathMap = new PathMap;
  if (dom) {
    events = model.__events = new EventDispatcher({
      onTrigger: function(name, listener, value, type, local, options) {
        var id, index, method, noRender, partial, path, property, _ref, _ref2;
        id = listener[0], method = listener[1], property = listener[2], partial = listener[3];
        path = pathMap.paths[name];
        if (method === 'propPolite' && local) {
          method = 'prop';
        }
        if (partial === '$inv') {
          value = !value;
        } else if (partial) {
          if (method === 'html' && type) {
            method = type;
            if (type === 'append') {
              path += '.' + (model.get(path).length - 1);
            } else if (type === 'insert') {
              _ref = value, index = _ref[0], value = _ref[1];
              path += '.' + index;
            } else if (type === 'remove') {
              noRender = true;
            } else if (type === 'move') {
              noRender = true;
              _ref2 = value, value = _ref2[0], property = _ref2[1];
            }
          } else if (method === 'attr') {
            value = null;
          }
          if (!noRender) {
            value = view.get(partial, value, null, null, path, id);
          }
        }
        return dom.update(id, method, options && options.ignore, value, property, index);
      }
    });
  } else {
    events = model.__events = new EventDispatcher;
  }
  events["__bind"] = events.bind;
  events.bind = function(name, listener) {
    return events["__bind"](pathMap.id(name), listener);
  };
  if (!dom) {
    return model;
  }
  model.on('set', function(_arg, local) {
    var path, value;
    path = _arg[0], value = _arg[1];
    return events.trigger(pathMap.id(path), value, 'html', local);
  });
  model.on('del', function(_arg, local) {
    var path;
    path = _arg[0];
    return events.trigger(pathMap.id(path), void 0, 'html', local);
  });
  model.on('push', function(_arg, local) {
    var id, path, value, values, _i, _len;
    path = _arg[0], values = 2 <= _arg.length ? __slice.call(_arg, 1) : [];
    id = pathMap.id(path);
    for (_i = 0, _len = values.length; _i < _len; _i++) {
      value = values[_i];
      events.trigger(id, value, 'append', local);
    }
  });
  refIndex = function(obj) {
    if (typeof obj === 'object') {
      return obj.index;
    } else {
      return obj;
    }
  };
  incrementMapItems = function(path, map, start, end, byNum) {
    var i, id, ids, itemPath, remainder, _results;
    _results = [];
    for (i = start; start <= end ? i <= end : i >= end; start <= end ? i++ : i--) {
      if (!(ids = map[i])) {
        continue;
      }
      _results.push((function() {
        var _results2;
        _results2 = [];
        for (id in ids) {
          remainder = ids[id];
          itemPath = pathMap.paths[id];
          delete pathMap.ids[itemPath];
          itemPath = path + '.' + (i + byNum) + remainder;
          pathMap.paths[id] = itemPath;
          _results2.push(pathMap.ids[itemPath] = +id);
        }
        return _results2;
      })());
    }
    return _results;
  };
  model.on('move', function(_arg, local) {
    var from, item, map, options, path, to;
    path = _arg[0], from = _arg[1], to = _arg[2], options = _arg[3];
    from = refIndex(from);
    to = refIndex(to);
    if (from === to) {
      return;
    }
    if (map = pathMap.arrays[path]) {
      incrementMapItems(path, map, from, from, to - from);
      if (from > to) {
        incrementMapItems(path, map, to, from - 1, 1);
      } else {
        incrementMapItems(path, map, from + 1, to, -1);
      }
      item = map.splice(from, 1)[0];
      map.splice(to, 0, item);
    }
    return events.trigger(pathMap.id(path), [from, to], 'move', local, options);
  });
  insert = function(path, index, values, local) {
    var howMany, i, id, ids, itemPath, len, map, remainder, value, _len;
    index = refIndex(index);
    if (map = pathMap.arrays[path]) {
      howMany = values.length;
      len = map.length;
      for (i = index; index <= len ? i < len : i > len; index <= len ? i++ : i--) {
        if (!(ids = map[i])) {
          continue;
        }
        for (id in ids) {
          remainder = ids[id];
          itemPath = pathMap.paths[id];
          delete pathMap.ids[itemPath];
          itemPath = path + '.' + (i + howMany) + remainder;
          pathMap.paths[id] = itemPath;
          pathMap.ids[itemPath] = +id;
        }
      }
      while (howMany--) {
        map.splice(index, 0, {});
      }
    }
    id = pathMap.id(path);
    for (i = 0, _len = values.length; i < _len; i++) {
      value = values[i];
      events.trigger(id, [index + i, value], 'insert', local);
    }
  };
  remove = function(path, start, howMany, local) {
    var end, i, id, ids, index, itemPath, len, map, remainder;
    start = refIndex(start);
    end = start + howMany;
    if (map = pathMap.arrays[path]) {
      len = map.length;
      for (i = start; start <= len ? i < len : i > len; start <= len ? i++ : i--) {
        if (!(ids = map[i])) {
          continue;
        }
        if (i < end) {
          for (id in ids) {
            itemPath = pathMap.paths[id];
            delete pathMap.ids[itemPath];
            delete pathMap.paths[id];
          }
        } else {
          for (id in ids) {
            remainder = ids[id];
            itemPath = pathMap.paths[id];
            delete pathMap.ids[itemPath];
            itemPath = path + '.' + (i - howMany) + remainder;
            pathMap.paths[id] = itemPath;
            pathMap.ids[itemPath] = +id;
          }
        }
      }
      map.splice(start, howMany);
    }
    id = pathMap.id(path);
    for (index = start; start <= end ? index < end : index > end; start <= end ? index++ : index--) {
      events.trigger(id, index, 'remove', local);
    }
  };
  model.on('unshift', function(_arg, local) {
    var path, values;
    path = _arg[0], values = 2 <= _arg.length ? __slice.call(_arg, 1) : [];
    return insert(path, 0, values, local);
  });
  model.on('insertBefore', function(_arg, local) {
    var index, path, value;
    path = _arg[0], index = _arg[1], value = _arg[2];
    return insert(path, index, [value], local);
  });
  model.on('insertAfter', function(_arg, local) {
    var index, path, value;
    path = _arg[0], index = _arg[1], value = _arg[2];
    return insert(path, index + 1, [value], local);
  });
  model.on('remove', function(_arg, local) {
    var howMany, path, start;
    path = _arg[0], start = _arg[1], howMany = _arg[2];
    return remove(path, start, howMany, local);
  });
  model.on('pop', function(_arg, local) {
    var path;
    path = _arg[0];
    return remove(path, model.get(path).length, 1, local);
  });
  model.on('shift', function(_arg, local) {
    var path;
    path = _arg[0];
    return remove(path, 0, 1, local);
  });
  model.on('splice', function(_arg, local) {
    var howMany, path, start, values;
    path = _arg[0], start = _arg[1], howMany = _arg[2], values = 4 <= _arg.length ? __slice.call(_arg, 3) : [];
    remove(path, start, howMany, local);
    return insert(path, index, values, local);
  });
  _ref = ['connected', 'canConnect'];
  _fn = function(event) {
    return model.on(event, function(value) {
      return events.trigger(pathMap.id(event), value);
    });
  };
  for (_i = 0, _len = _ref.length; _i < _len; _i++) {
    event = _ref[_i];
    _fn(event);
  }
  return model;
};